"""Test end-to-end but mock out Bitfount Hub API (as well as message service)."""
import asyncio
import copy
import logging
import multiprocessing
from multiprocessing import current_process
from multiprocessing.managers import DictProxy, SyncManager
import os
from pathlib import Path
from queue import Queue
import time
from typing import Any, Callable, Dict, Generator, Iterable, Type, Union, cast
from unittest.mock import Mock

from _pytest.monkeypatch import MonkeyPatch
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from flask import Flask, url_for
import pytest
from pytest import fixture
from pytest_flask.live_server import LiveServer
from pytest_mock import MockerFixture
from pytorch_lightning.loggers import TensorBoardLogger

import bitfount
from bitfount import PodRegistrationError
from bitfount.federated.modeller import _Modeller
from bitfount.federated.transport.base_transport import _run_func_and_listen_to_mailbox
from bitfount.federated.transport.message_service import _LargeObjectRequestHandler
from bitfount.hub.authentication_flow import BitfountSession
from bitfount.runners.config_schemas import ModellerConfig, PodConfig
from bitfount.runners.modeller_runner import setup_modeller_from_config
from bitfount.runners.pod_runner import setup_pod_from_config
from bitfount.types import (
    _JSONDict,
    _S3PresignedPOSTFields,
    _S3PresignedPOSTURL,
    _S3PresignedURL,
)
from tests.integration import CONFIG_DIR
from tests.integration.utils import (
    get_caplog_records,
    load_modeller_config,
    load_pod_config,
    pod_start,
    run_modeller_process,
)
from tests.utils.helper import backend_test, end_to_end_mocks_test
from tests.utils.mocks import GRPCStubMock

STD_TIMEOUT: int = 3 * 60  # 3 minutes
EXT_TIMEOUT: int = 8 * 60  # 8 minutes

PROCESS_TIMEOUT: int = 6 * 60  # 6 minutes


logger = logging.getLogger(__name__)


class MockBitfountSession(BitfountSession):
    """A mock implementation of BitfountSession which avoids authentication."""

    def __init__(self, access_token: str, username: str, **kwargs: Any):
        self._faked_access_token = access_token
        super().__init__(username=username, **kwargs)

    @property
    def access_token(self) -> str:
        """The supplied access token to provide to the message service."""
        return self._faked_access_token

    def authenticate(self) -> None:
        """Same as BitfountSession but doesn't authenticate anything."""
        pass

    @property
    def username(self) -> str:
        """Returns the username the authenticated user would have."""
        return self._username


@end_to_end_mocks_test
class TestEndToEnd:
    """Tests end-to-end functionality using mocks."""

    @fixture(autouse=True)
    def mock_am(
        self, monkeypatch: MonkeyPatch, access_manager_public_key: RSAPublicKey
    ) -> None:
        """Mock get_access_manager_key."""
        import bitfount.hub.api as api_

        monkeypatch.setattr(
            api_.BitfountAM,
            "get_access_manager_key",
            lambda x: access_manager_public_key,
        )

    @fixture(autouse=True)
    def mock_logger_save_dir(self, monkeypatch: MonkeyPatch, tmp_path: Path) -> None:
        """Mock pytorch logger."""
        from bitfount.backends.pytorch.models.base_models import BasePyTorchModel

        monkeypatch.setattr(
            BasePyTorchModel,
            "_get_logger_from_config",
            lambda *args, **kwargs: TensorBoardLogger(
                str(tmp_path),
                name="mock_e2e_tests_model_name",
            ),
        )

    @fixture
    def pod_config(self, census_income_data: Path) -> PodConfig:
        """Pod config to use for tests."""
        pod_config = load_pod_config(CONFIG_DIR / "pod_1.yaml", census_income_data)
        pod_config.approved_pods = [f"{pod_config.pod_id}2"]
        pod_config.hub.url = url_for("index", _external=True)
        return pod_config

    @fixture
    def modeller_config(
        self, modeller_private_key: RSAPrivateKey
    ) -> Generator[ModellerConfig, None, None]:
        """Modeller config to use for tests."""
        try:
            modeller_config = load_modeller_config(
                CONFIG_DIR / "modeller.yaml",
                private_key=modeller_private_key,
            )
            modeller_config.message_service.url = "http://hub.bitfount.notarealms.com"
            modeller_config.pods.identifiers = ["e2e_provider_1/pod"]
            modeller_config.hub.url = url_for("index", _external=True)

            yield modeller_config
        finally:
            try:
                # The unbound variable issue is handled by the NameError catching below.
                # noinspection PyUnboundLocalVariable
                private_key_file: Path = cast(
                    Path, modeller_config.modeller.private_key_file
                )
                os.remove(private_key_file)
            except FileNotFoundError:
                pass
            except NameError:
                # Means the modeller_config variable didn't even get created;
                # another exception has occurred that has booted us out of the
                # `try` block before it even got to there. We pass on NameError
                # to avoid masking whatever that underlying exception was.
                pass

    @fixture
    def bitfount_session_factory(
        self, tmp_path: Path
    ) -> Callable[[str, str], MockBitfountSession]:
        """Mock bitfount session factory."""

        def _create_mock_bitfount_session(
            username: str, access_token: str
        ) -> MockBitfountSession:
            session = MockBitfountSession(access_token=access_token, username=username)
            session.user_storage_path = tmp_path / username
            session.token_file = session.user_storage_path / ".token"
            return session

        return _create_mock_bitfount_session

    @fixture
    def apply_mock_large_object_interactions(
        self, mocker: MockerFixture
    ) -> Callable[[SyncManager], None]:
        """Returns a function to apply a mock to LargeObjectRequestHandler.

        A function is needed due to the involved Manager.dict() call; the manager
        must be provided at runtime.
        """

        def _mock_large_object_interactions(manager: SyncManager) -> None:
            """Mock out LargeObjectRequestHandler methods.

            Mocks out LargeObjectRequestHandler.upload_large_object() and
            LargeObjectRequestHandler.get_large_object_from_url() to use local storage.
            """
            # Mock out message service functions
            large_object_storage: DictProxy[
                Union[_S3PresignedURL, _S3PresignedPOSTURL], bytes
            ] = manager.dict()

            # upload_fields is unused below
            # noinspection PyUnusedLocal
            def store_large_object(
                upload_url: _S3PresignedPOSTURL,
                upload_fields: _S3PresignedPOSTFields,
                large_object: bytes,
            ) -> None:
                """Mocks uploading to the URL.

                Args:
                    upload_url: URL to "upload" to.
                    upload_fields: Unused. Fields to upload with.
                    large_object: Object to store.
                """
                large_object_storage[upload_url] = large_object

            def get_large_object(url: _S3PresignedURL) -> bytes:
                """Mock downloading the object from the download URL.

                Args:
                    url: The download URL provided.

                Returns: The retrieved object/message
                """
                # This is just to simulate the fact that
                # the uploader and the downloader have used different URLs,
                # so we have to tweak one of them to find the object
                fixed_url: _S3PresignedPOSTURL = cast(
                    _S3PresignedPOSTURL, url.replace("download", "upload")
                )
                retrieved_object = large_object_storage[fixed_url]
                return retrieved_object

            mocker.patch.object(
                _LargeObjectRequestHandler,
                "upload_large_object",
                new=store_large_object,
            )
            mocker.patch.object(
                _LargeObjectRequestHandler,
                "get_large_object_from_url",
                new=get_large_object,
            )

        return _mock_large_object_interactions

    @staticmethod
    def _modeller_send_task_requests(
        modeller: _Modeller,
        pod_identifiers: Iterable[str],
    ) -> None:
        """Helper for sending Modeller task requests in a separate process.

        Fails tests if there are any errors.
        """
        # As this is used in multiprocessing, _modeller_send_task_requests() has
        # to be a def function. So we wrap the desired methods in an async def
        # and use asyncio.run() to simulate async def.
        async def _run_modeller() -> None:
            """Asyncronous modeller runner to process and run task requests."""
            modeller_mailbox = await modeller._send_task_requests(pod_identifiers)
            await _run_func_and_listen_to_mailbox(
                modeller_mailbox.process_task_request_responses(), modeller_mailbox
            )

        try:
            asyncio.run(_run_modeller())
        except Exception as e:
            print(f"Caught exception in {current_process().name} process: {e}")
            raise e

    @pytest.mark.parametrize(
        argnames=("cli_mode", "expected_exc"),
        argvalues=((True, SystemExit), (False, PodRegistrationError)),
    )
    @pytest.mark.parametrize("app", [(True, False, False)], indirect=True)
    @pytest.mark.timeout(STD_TIMEOUT)
    def test_pod_setup_raises_pod_registration_error(
        self,
        app: Flask,
        bitfount_session_factory: Callable[[str, str], MockBitfountSession],
        cli_mode: bool,
        expected_exc: Type[BaseException],
        live_server: LiveServer,
        mock_bitfount_session: Mock,
        monkeypatch: MonkeyPatch,
        pod_config: PodConfig,
    ) -> None:
        """Tests handling when pod raises registration error."""
        # Set _BITFOUNT_CLI_MODE config variable to elicit different error handling
        monkeypatch.setattr(bitfount.config, "_BITFOUNT_CLI_MODE", cli_mode)

        username = pod_config.username
        # Override default return_value with bitfount_session_factory() one
        mock_bitfount_session.return_value = bitfount_session_factory(username, "pod")

        with pytest.raises(expected_exc):
            setup_pod_from_config(pod_config)

    @pytest.mark.timeout(STD_TIMEOUT)
    def test_pod_starts_successfully(
        self,
        bitfount_session_factory: Callable[[str, str], MockBitfountSession],
        caplog_queue: Queue,
        live_server: LiveServer,  # automatically finds app fixture
        mock_bitfount_session: Mock,
        mock_grpc_insecure_channel: Mock,
        mock_message_service_stub: Mock,
        mock_s3_data_upload_in_api_module: Mock,
        pod_config: PodConfig,
    ) -> None:
        """Starts Pod listening in a subprocess.

        Sends SIGINT to the Pod after 10s to stop listening
        (Otherwise this test will hang indefinitely)
        """
        with multiprocessing.Manager() as manager:
            pod_token = "pod"
            username = pod_config.username
            mock_message_service_stub.return_value = GRPCStubMock(
                {pod_token: pod_config.pod_id}, manager
            )
            # Override default return_value with bitfount_session_factory() one
            mock_bitfount_session.return_value = bitfount_session_factory(
                username, pod_token
            )
            pod = setup_pod_from_config(pod_config)

            p = multiprocessing.Process(
                target=pod_start, name="Pod_Runner", args=(pod,)
            )
            p.start()
            time.sleep(10)
            p.terminate()
            p.join()

            caplog_records = get_caplog_records(caplog_queue)

            assert "INFO" in caplog_records
            assert "ERROR" not in caplog_records
            assert "CRITICAL" not in caplog_records

            assert p.exitcode is not None and p.exitcode <= 0

    @pytest.mark.skip("[BIT-1669] Fix e2e tests")
    @backend_test
    @pytest.mark.parametrize("app", [(False, True, False)], indirect=True)
    @pytest.mark.timeout(STD_TIMEOUT)
    async def test_modeller_send_task_request_rejected_by_pod(
        self,
        app: Flask,
        bitfount_session_factory: Callable[[str, str], MockBitfountSession],
        caplog_queue: Queue,
        live_server: LiveServer,
        mock_bitfount_session: Mock,
        mock_grpc_insecure_channel: Mock,
        mock_message_service_stub: Mock,
        mock_s3_data_upload_in_api_module: Mock,
        mocker: MockerFixture,
        modeller_config: ModellerConfig,
        pod_config: PodConfig,
    ) -> None:
        """Tests Modeller handling if task requests are rejected by pods."""
        try:
            with multiprocessing.Manager() as manager:
                modeller_token = "modeller_token"
                pod_token = "pod_token"

                # Mock out GRPC interactions
                mock_stub = GRPCStubMock(
                    {
                        modeller_token: modeller_config.modeller.username,
                        pod_token: pod_config.pod_id,
                    },
                    manager,
                )
                mock_message_service_stub.return_value = mock_stub

                # Ensure only testing against one pod
                modeller_config.pods.identifiers = [pod_config.pod_id]

                # Mock out the sessions used for authentication
                pod_username = pod_config.username
                modeller_username = modeller_config.modeller.username
                # Override default return_value with side_effect
                # bitfount_session_factory() one. side_effect takes precedence
                # over return_value.
                mock_bitfount_session.side_effect = [
                    bitfount_session_factory(pod_username, pod_token),
                    bitfount_session_factory(modeller_username, modeller_token),
                ]
                # Setup pod and modeller
                pod = setup_pod_from_config(pod_config)
                mocker.patch(
                    "bitfount.runners.modeller_runner.get_pod_schema",
                    return_value=pod.schema,
                )

                modeller, pod_identifiers = setup_modeller_from_config(modeller_config)
                print(f"{modeller._identity_verification_method}")

                # Start pod, give time to spin up
                pod_process = multiprocessing.Process(
                    target=pod_start, name="Pod_Runner", args=(pod,)
                )
                pod_process.start()
                time.sleep(2)

                # Start modeller
                mocker.patch("bitfount.hub.helper._save_key_to_key_store")
                modeller_process = multiprocessing.Process(
                    target=self._modeller_send_task_requests,
                    name="Modeller_Runner",
                    args=(modeller, pod_identifiers),
                )
                modeller_process.start()

                # Give time for processing to happen then stop processes
                time.sleep(20)
                for proc in [pod_process, modeller_process]:
                    proc.terminate()
                    proc.join()

                # Check conditions in log messages match rejection of task
                caplog_records = get_caplog_records(caplog_queue)
                assert "INFO" in caplog_records
                assert (
                    "Task request from e2e_modeller rejected. "
                    "Protocol not approved" in caplog_records["INFO"]
                )
                assert "ERROR" in caplog_records
                assert (
                    f"Received rejection from {pod_config.pod_id}. "
                    f"Protocol not approved" in caplog_records["ERROR"]
                )
                assert "CRITICAL" not in caplog_records

                # Check that things shutdown correctly
                assert (
                    modeller_process.exitcode is not None
                    and modeller_process.exitcode <= 0
                )
                assert pod_process.exitcode is not None and pod_process.exitcode <= 0
        finally:
            # noinspection PyBroadException
            try:
                # noinspection PyUnboundLocalVariable
                modeller_process.terminate()
            except Exception:
                pass

            # noinspection PyBroadException
            try:
                # noinspection PyUnboundLocalVariable
                pod_process.terminate()
            except Exception:
                pass

    @fixture
    def mock_schema_upload_download(self, mocker: MockerFixture) -> None:
        """Mocks out schema upload and download in the api module."""
        schema_store: Dict[str, _JSONDict] = {}

        def _upload(
            upload_url: _S3PresignedPOSTURL,
            presigned_fields: _S3PresignedPOSTFields,
            data: _JSONDict,
        ) -> None:
            pod_name = upload_url.split("/")[-1]
            # Error out if upload already used
            if pod_name in schema_store:
                raise ValueError("Cannot reuse schema upload URL")
            schema_store[pod_name] = data
            logger.info(f"Schema uploaded for {pod_name}")

        def _download(download_url: _S3PresignedURL) -> _JSONDict:
            pod_name = download_url.split("/")[-1]
            logger.info(f"Schema downloaded for {pod_name}")
            return schema_store[pod_name]

        mocker.patch(
            "bitfount.hub.api._upload_data_to_s3", autospec=True, side_effect=_upload
        )
        mocker.patch(
            "bitfount.hub.api._download_data_from_s3",
            autospec=True,
            side_effect=_download,
        )

    @pytest.mark.skip("[BIT-1669] Fix e2e tests")
    @backend_test
    @pytest.mark.timeout(EXT_TIMEOUT)
    @pytest.mark.parametrize("use_local_storage", [True, False])
    async def test_full_training_end_to_end_successfully(
        self,
        apply_mock_large_object_interactions: Callable[[SyncManager], None],
        bitfount_session_factory: Callable[[str, str], MockBitfountSession],
        caplog_queue: Queue,
        live_server: LiveServer,  # automatically finds app fixture
        mock_bitfount_session: Mock,
        mock_grpc_insecure_channel: Mock,
        mock_message_service_stub: Mock,
        mock_schema_upload_download: None,
        mocker: MockerFixture,
        modeller_config: ModellerConfig,
        pod_config: PodConfig,
        tmp_path: Path,
        use_local_storage: bool,
    ) -> None:
        """Tests that training works end to end.

        Parameterised to run by sending messages directly, and then sending messages
        as file paths.
        """
        try:
            with multiprocessing.Manager() as manager:
                pod1_config = pod_config

                # Determine if we are using file paths for messages and set it up
                if use_local_storage:
                    pod1_config.message_service.use_local_storage = True
                    modeller_config.message_service.use_local_storage = True

                # Create second pod config from first
                second_pod_owner = "e2e_provider_2"
                pod2_config = copy.deepcopy(pod1_config)
                pod2_config.name += "2"
                pod2_config.username = second_pod_owner

                # Patch out GRPC interactions
                pod1_token = "pod1"
                modeller_token = "modeller_token"
                pod2_token = "pod2"
                mock_stub = GRPCStubMock(
                    {
                        modeller_token: modeller_config.modeller.username,
                        pod1_token: pod1_config.pod_id,
                        pod2_token: pod2_config.pod_id,
                    },
                    manager,
                    get_message_timeout=5,
                )
                mock_message_service_stub.return_value = mock_stub

                # Mock out large object handling in message service
                apply_mock_large_object_interactions(manager)

                # Tie pod configs together
                pod1_config.approved_pods = [pod2_config.pod_id]
                pod2_config.approved_pods = [pod1_config.pod_id]
                modeller_config.pods.identifiers = [
                    pod1_config.pod_id,
                    pod2_config.pod_id,
                ]

                # Generate mocked sessions for modeller and pods
                modeller_username = modeller_config.modeller.username
                pod1_username = pod1_config.username
                pod2_username = second_pod_owner
                # Override default return_value with side_effect
                # bitfount_session_factory() one. side_effect takes precedence
                # over return_value.
                mock_bitfount_session.side_effect = [
                    # NOTE: Order is important; this is the order they will be
                    # setup in below.
                    bitfount_session_factory(pod1_username, pod1_token),
                    bitfount_session_factory(pod2_username, pod2_token),
                    bitfount_session_factory(modeller_username, modeller_token),
                ]

                # Setup pods and modeller (Note: order is important)
                pod1 = setup_pod_from_config(pod1_config)
                pod2 = setup_pod_from_config(pod2_config)

                modeller, pod_identifiers = setup_modeller_from_config(modeller_config)
                model_out = tmp_path / "model.out"

                # Start pods and wait to spin up
                pod_1_process = multiprocessing.Process(
                    target=pod_start, name="Pod_1_Runner", args=(pod1,)
                )
                pod_1_process.start()
                time.sleep(0.5)
                pod_2_process = multiprocessing.Process(
                    target=pod_start, name="Pod_2_Runner", args=(pod2,)
                )
                pod_2_process.start()
                time.sleep(0.5)

                # Start modeller
                mocker.patch("bitfount.hub.helper._save_key_to_key_store")
                modeller_process = multiprocessing.Process(
                    target=run_modeller_process,
                    name="MM_1",
                    args=(
                        modeller,
                        pod_identifiers,
                        model_out,
                    ),
                )
                modeller_process.start()

                # TODO: [BIT-350] instead of relying on a timeout - replace this
                #       with a call that waits for the process to finish before
                #       terminating. multiprocessing join() method currently hangs
                #       when using tox (but not pytest individually).
                # Wait for at most PROCESS_TIMEOUT seconds, checking every 30 seconds.
                for i in range(int(PROCESS_TIMEOUT / 30)):
                    time.sleep(30)
                    logger.info(f"{(i+1)*30} seconds...")
                    # the modeller is the only one that will terminate freely
                    # so check that.
                    if not modeller_process.is_alive():
                        break

                # End all processes, regardless of state
                for proc in [modeller_process, pod_1_process, pod_2_process]:
                    proc.terminate()
                    proc.join()

                caplog_records = get_caplog_records(caplog_queue)
                assert "INFO" in caplog_records
                assert "CRITICAL" not in caplog_records
                assert "ERROR" not in caplog_records

                info_logs = " ".join(caplog_records["INFO"])
                assert "Task complete" in info_logs
                assert "Validation Metrics" in info_logs

                assert (
                    modeller_process.exitcode is not None
                    and modeller_process.exitcode <= 0
                )
                assert (
                    pod_1_process.exitcode is not None and pod_1_process.exitcode <= 0
                )
                assert (
                    pod_2_process.exitcode is not None and pod_2_process.exitcode <= 0
                )
        finally:
            # Stop all processes
            # noinspection PyBroadException
            try:
                # noinspection PyUnboundLocalVariable
                modeller_process.terminate()
            except Exception:
                pass
            # noinspection PyBroadException
            try:
                # noinspection PyUnboundLocalVariable
                pod_1_process.terminate()
            except Exception:
                pass
            # noinspection PyBroadException
            try:
                # noinspection PyUnboundLocalVariable
                pod_2_process.terminate()
            except Exception:
                pass



"""PyTorch implementations for the Bitfount framework."""

from typing import List

from bitfount.backends.pytorch.data.dataloaders import (
    DEFAULT_BUFFER_SIZE,
    PyTorchBitfountDataLoader,
    PyTorchIterableBitfountDataLoader,
)
from bitfount.backends.pytorch.federated.shim import PyTorchBackendTensorShim
from bitfount.backends.pytorch.loss import SoftDiceLoss, soft_dice_loss
from bitfount.backends.pytorch.models.base_models import PyTorchClassifierMixIn
from bitfount.backends.pytorch.models.bitfount_model import PyTorchBitfountModel
from bitfount.backends.pytorch.models.models import (
    PyTorchImageClassifier,
    PyTorchTabularClassifier,
    TabNetClassifier,
)
from bitfount.backends.pytorch.models.torch_functions.mish import Mish
from bitfount.backends.pytorch.utils import LoggerType

__all__: List[str] = [
    "DEFAULT_BUFFER_SIZE",
    "LoggerType",
    "Mish",
    "PyTorchBackendTensorShim",
    "PyTorchBitfountDataLoader",
    "PyTorchBitfountModel",
    "PyTorchClassifierMixIn",
    "PyTorchImageClassifier",
    "PyTorchIterableBitfountDataLoader",
    "PyTorchTabularClassifier",
    "soft_dice_loss",
    "SoftDiceLoss",
    "TabNetClassifier",
]

# See top level `__init__.py` for an explanation
__pdoc__ = {}
for _obj in __all__:
    __pdoc__[_obj] = False



"""Tests for the base transformation classes."""
from marshmallow import Schema as MarshmallowSchema
from marshmallow import ValidationError
import pytest

from bitfount.transformations.base_transformation import (
    Transformation,
    _TransformationSchema,
)
from bitfount.transformations.exceptions import TransformationRegistryError
from tests.bitfount.transformations.transformation_test_helpers import gen_name_fails
from tests.utils.helper import unit_test


@unit_test
class TestTransformationRegistry:
    """Tests for the functionality of the transformation registry."""

    def test_transformations_are_registered(self, empty_registry: None) -> None:
        """Tests that concrete transformation classes are correctly registered."""
        from bitfount.transformations.base_transformation import (
            TRANSFORMATION_REGISTRY,
            Transformation,
        )

        class Test(Transformation):
            _registry_name = "test"

        class Test_2(Transformation):
            _registry_name = "test_2"

        assert TRANSFORMATION_REGISTRY == {"test": Test, "test_2": Test_2}

    def test_transformation_registry_names_unique(self, empty_registry: None) -> None:
        """Tests that transformation registry names are unique.

        Tests exception is raised if multiple concrete transformations attempt to
        register with same name.
        """
        from bitfount.transformations.base_transformation import Transformation

        class _Test(Transformation):
            _registry_name = "test"

        with pytest.raises(
            TransformationRegistryError,
            match='A transformation is already registered with name "test"',
        ):

            class _Test_2(Transformation):
                _registry_name = "test"

    def test_non_concrete_transformations_not_registered(
        self, empty_registry: None
    ) -> None:
        """Tests non-concrete classes aren't registered."""
        from bitfount.transformations.base_transformation import (
            TRANSFORMATION_REGISTRY,
            Transformation,
        )

        class _AbstractTest(Transformation):
            pass

        assert TRANSFORMATION_REGISTRY == {}


@unit_test
class TestTransformation:
    """Tests the base transformation class."""

    def test_schema_loads(self) -> None:
        """Tests schema load works correctly."""
        s = Transformation.schema()
        t = s.loads(
            """
            name: test
            output: True
            """
        )
        assert t.name == "test"
        assert t.output is True

    def test_schema_load_fails_without_name(self) -> None:
        """Tests schema loading fails without name field."""
        s = Transformation.schema()
        with pytest.raises(
            TransformationRegistryError,
            match="Transformation .* isn't registered; can't generate name",
        ):
            s.loads(
                """
                output: True
                """
            )

    def test_schema_load_succeeds_without_output(self) -> None:
        """Tests schema loads successfully without output field."""
        s = Transformation.schema()
        t = s.loads(
            """
            name: test
            """
        )
        assert t.name == "test"
        assert t.output is False

    def test_schema_load_fails_with_wrong_type_name(self) -> None:
        """Tests schema load fails if non-str name field."""
        s = Transformation.schema()
        with pytest.raises(ValidationError):
            s.loads(
                """
                name: 2
                output: True
                """
            )

    def test_schema_load_fails_with_wrong_type_output(self) -> None:
        """Tests schema load fails if non-bool output field."""
        s = Transformation.schema()
        with pytest.raises(ValidationError):
            s.loads(
                """
                name: test
                output: hello
                """
            )

    def test__gen_name_fails(self) -> None:
        """Tests that name generation fails."""
        gen_name_fails(Transformation)

    def test_custom_schema_loads(self) -> None:
        """Tests that if a custom schema is defined, it is the one loaded."""

        class CustomTransformation(Transformation):
            class _Schema(_TransformationSchema):
                pass

        s = CustomTransformation.schema()
        assert isinstance(s, CustomTransformation._Schema)

    def test_custom_schema_must_subclass_transformation_schema(self) -> None:
        """Tests that error raised if custom schema isn't a TransformationSchema."""

        class CustomTransformation(Transformation):
            class _Schema(MarshmallowSchema):
                pass

        with pytest.raises(TypeError, match="must be a TransformationSchema instance."):
            CustomTransformation.schema()



"""Tests federated helper.py."""
import copy
from typing import Optional, cast
from unittest.mock import create_autospec

from _pytest.logging import LogCaptureFixture
from _pytest.monkeypatch import MonkeyPatch
import pytest
from pytest import fixture
from pytest_lazyfixture import lazy_fixture
from pytest_mock import MockerFixture

from bitfount.data.schema import BitfountSchema, BitfountSchemaError
from bitfount.federated.aggregators.aggregator import Aggregator
from bitfount.federated.aggregators.secure import SecureAggregator
from bitfount.federated.exceptions import PodNameError
from bitfount.federated.helper import (
    _check_and_update_pod_ids,
    _create_aggregator,
    _create_and_connect_pod_mailbox,
    _create_message_service,
    _get_idp_url,
    combine_pod_schemas,
)
from bitfount.federated.transport.config import (
    _DEV_MESSAGE_SERVICE_PORT,
    _DEV_MESSAGE_SERVICE_TLS,
    _DEV_MESSAGE_SERVICE_URL,
    _STAGING_MESSAGE_SERVICE_URL,
    PRODUCTION_MESSAGE_SERVICE_URL,
    MessageServiceConfig,
)
from bitfount.federated.transport.message_service import _MessageService
from bitfount.federated.transport.pod_transport import _PodMailbox
from bitfount.hub.authentication_flow import BitfountSession
from bitfount.hub.types import _DEV_IDP_URL, _PRODUCTION_IDP_URL, _STAGING_IDP_URL
from tests.utils.helper import create_schema, get_warning_logs, unit_test


@unit_test
class TestHelperFunctions:
    """Test federated helper functions."""

    @fixture
    def test_url(self) -> str:
        """Test URL."""
        return "not.a.real.url.com"

    @fixture
    def schema(self) -> BitfountSchema:
        """Creates a schema fixture."""
        return create_schema(classification=True)

    def test_create_insecure_aggregator_successfully(self) -> None:
        """Tests create_aggregator function without secure aggregation."""
        aggregator = _create_aggregator(secure_aggregation=False)
        assert isinstance(aggregator, Aggregator)
        assert not isinstance(aggregator, SecureAggregator)

    def test_create_secure_aggregator_successfully(self) -> None:
        """Tests create_aggregator function with secure aggregation."""
        aggregator = _create_aggregator(secure_aggregation=True)
        assert isinstance(aggregator, SecureAggregator)

    @pytest.mark.parametrize(
        "environment, expected_config, input_config",
        [
            (
                "production",
                MessageServiceConfig(url=PRODUCTION_MESSAGE_SERVICE_URL),
                None,
            ),
            ("staging", MessageServiceConfig(url=_STAGING_MESSAGE_SERVICE_URL), None),
            (
                "dev",
                MessageServiceConfig(
                    url=_DEV_MESSAGE_SERVICE_URL,
                    port=_DEV_MESSAGE_SERVICE_PORT,
                    tls=_DEV_MESSAGE_SERVICE_TLS,
                ),
                None,
            ),
            (
                None,
                MessageServiceConfig(url=cast(str, lazy_fixture("test_url"))),
                MessageServiceConfig(url=cast(str, lazy_fixture("test_url"))),
            ),
        ],
        indirect=["environment"],
    )
    def test_create_message_service_successfully(
        self,
        environment: None,
        expected_config: MessageServiceConfig,
        input_config: Optional[MessageServiceConfig],
    ) -> None:
        """Tests create_message_service method works.

        Ensures correct config is passed through to the created message service.
        """
        message_service = _create_message_service(
            session=create_autospec(BitfountSession, instance=True),
            ms_config=input_config,
        )
        assert isinstance(message_service, _MessageService)
        assert message_service._config == expected_config

    def test_create_message_service_issues_warning_for_local_storage(
        self, caplog: LogCaptureFixture
    ) -> None:
        """Tests create_message_service issues warning if using local storage."""
        ms_config = MessageServiceConfig(use_local_storage=True)
        message_service = _create_message_service(
            session=create_autospec(BitfountSession, instance=True),
            ms_config=ms_config,
        )
        assert isinstance(message_service, _MessageService)

        # Check for warning message; should be thing logged
        warning_logs = get_warning_logs(caplog)
        assert (
            "Messages will contain local file references. "
            "Ensure all pods have access to your local file system. "
            "Otherwise your task will hang." in warning_logs
        )

    async def test_create_and_connect_pod_mailbox_successfully(
        self, mocker: MockerFixture
    ) -> None:
        """Tests create_and_connect_pod_mailbox method works."""
        # We know that `create_message_service` is already tested so we mock out
        # its response and check it is called correctly.
        mock_create_message_service = mocker.patch(
            "bitfount.federated.helper._create_message_service"
        )

        # We also need to mock out the PodMailbox.connect_pod() class method.
        mock_connect_pod = mocker.patch.object(
            _PodMailbox, "connect_pod", autospec=True
        )
        mock_connect_pod.return_value = create_autospec(_PodMailbox, instance=True)

        # Create mocks for args we don't care about
        pod_name = "not-a-real-pod"
        mock_session = create_autospec(BitfountSession, instance=True)
        mock_ms_config = create_autospec(MessageServiceConfig, instance=True)

        mailbox = await _create_and_connect_pod_mailbox(
            pod_name=pod_name,
            session=mock_session,
            ms_config=mock_ms_config,
        )

        assert isinstance(mailbox, _PodMailbox)
        mock_connect_pod.assert_called_once_with(
            pod_name=pod_name, message_service=mock_create_message_service()
        )

    def test__check_and_update_pod_ids(self, mocker: MockerFixture) -> None:
        """Tests that the pod identifiers are updated correctly."""
        mock_hub = mocker.patch("bitfount.hub.api.BitfountHub")
        mock_hub.username = "my-username"
        pod_id_1 = "my-pod"
        pod_id_2 = "your-username/your-pod"
        pod_ids = _check_and_update_pod_ids([pod_id_1, pod_id_2], mock_hub)
        assert pod_ids == ["my-username/my-pod", "your-username/your-pod"]

    @pytest.mark.parametrize(
        "invalid_pod_id",
        ["Invalid", "INVALID", "invalid-", "-invalid", "invalid--invalid"],
    )
    def test__check_and_update_pod_ids_errors(
        self, invalid_pod_id: str, mocker: MockerFixture
    ) -> None:
        """Tests that the pod identifiers error correctly."""
        mock_hub = mocker.patch("bitfount.hub.api.BitfountHub")
        mock_hub.username = "my-username"

        with pytest.raises(PodNameError):
            _check_and_update_pod_ids([invalid_pod_id], mock_hub)

    def test_idp_url_gets_staging_url(self, monkeypatch: MonkeyPatch) -> None:
        """Tests that the idp_url is loaded based on staging environment."""
        monkeypatch.setenv("BITFOUNT_ENVIRONMENT", "staging")
        idp_url = _get_idp_url()
        assert idp_url == _STAGING_IDP_URL

    def test_idp_url_gets_prod_url(self, monkeypatch: MonkeyPatch) -> None:
        """Tests that the idp_url is loaded based on staging environment."""
        monkeypatch.setenv("BITFOUNT_ENVIRONMENT", "production")
        idp_url = _get_idp_url()
        assert idp_url == _PRODUCTION_IDP_URL

    def test_idp_url_gets_dev_url(self, monkeypatch: MonkeyPatch) -> None:
        """Tests that the idp_url is loaded based on dev environment."""
        monkeypatch.setenv("BITFOUNT_ENVIRONMENT", "dev")
        idp_url = _get_idp_url()
        assert idp_url == _DEV_IDP_URL

    def test_combine_pod_schemas_raises_error_if_there_are_duplicated_table_names(
        self, mocker: MockerFixture, schema: BitfountSchema
    ) -> None:
        """Tests that schemas can't be combined if they have the same table names."""
        schema_2 = copy.deepcopy(schema)
        mocker.patch(
            "bitfount.federated.helper.get_pod_schema", side_effect=[schema, schema_2]
        )
        with pytest.raises(BitfountSchemaError):
            combine_pod_schemas(["pod_1", "pod_2"])

    def test_combine_pod_schemas_correctly(
        self, mocker: MockerFixture, schema: BitfountSchema
    ) -> None:
        """Tests that schemas can be combined correctly."""
        schema_2 = copy.deepcopy(schema)
        schema_2.tables[0].name = "table_2"
        mocker.patch(
            "bitfount.federated.helper.get_pod_schema", side_effect=[schema, schema_2]
        )
        combined_schema = combine_pod_schemas(["pod_1", "pod_2"])
        assert isinstance(combined_schema, BitfountSchema)
        assert len(combined_schema.tables) == 2
        assert combined_schema.table_names == schema.table_names + schema_2.table_names



"""References and constants for transformations.

Contains regular expressions and methods for matching transformation and
column references.
"""
import re
from typing import cast

from bitfount.transformations.exceptions import (
    IncorrectReferenceError,
    NotColumnReferenceError,
    NotTransformationReferenceError,
)

_TRANSFORMATION_REFERENCE = re.compile("(?:t|tran|transformation):(.*)", re.IGNORECASE)
_COLUMN_REFERENCE = re.compile("(?:c|col|column):(.*)", re.IGNORECASE)


def _extract_ref_regex(s: str, ref_regex: re.Pattern) -> str:
    """Extracts a reference based on a pattern.

    Args:
        s: The potential reference.
        ref_regex: The regex pattern to match against.

    Returns:
        The name being referenced.

    Raises:
        IncorrectReferenceError:
            If no reference was found according to the pattern.
        TypeError:
            If `s` type cannot be matched.
    """
    match = ref_regex.fullmatch(s)
    if match:
        return cast(str, match.group(1))
    else:
        raise IncorrectReferenceError


def _extract_transformation_ref(t_str: str) -> str:
    """Extracts the transformation name being referenced if possible.

    Args:
        t_str: The potential transformation reference.

    Returns:
        The name of the transformation being referenced.

    Raises:
        NotTransformationReferenceError:
            Raised if this is not a transformation reference by type or
            content.
    """
    try:
        return _extract_ref_regex(t_str, _TRANSFORMATION_REFERENCE)
    except IncorrectReferenceError:
        raise NotTransformationReferenceError(
            "Incorrect format for transformation reference; transformation references"
            'should start with "t:", "tran:" or "transformation:"'
        )
    except TypeError:
        raise NotTransformationReferenceError(
            f"Incorrect type for transformation reference; "
            f"expected str, got {type(t_str)}"
        )


def _extract_col_ref(col_str: str) -> str:
    """Extracts the column name being referenced if possible.

    Args:
        col_str: The potential column reference.

    Returns:
        The name of the column being referenced.

    Raises:
        NotColumnReferenceError:
            Raised if this is not a column reference by type or content.
    """
    try:
        return _extract_ref_regex(col_str, _COLUMN_REFERENCE)
    except IncorrectReferenceError:
        raise NotColumnReferenceError(
            "Incorrect format for column reference; column references should "
            'start with "c:", "col:" or "column:"'
        )
    except TypeError:
        raise NotColumnReferenceError(
            f"Incorrect type for column reference; expected str, got {type(col_str)}"
        )


def _extract_ref(r_str: str) -> str:
    """Extracts the column or transformation name being referenced if possible.

    Args:
        r_str: The potential column or transformation reference.

    Returns:
        The name of the column or transformation being referenced.

    Raises:
        IncorrectReferenceError:
            Raised if this is not a column or transformation reference by type
            or content.
    """
    # Try transformation reference first
    try:
        return _extract_transformation_ref(r_str)
    except NotTransformationReferenceError:
        pass
    # Otherwise try column reference
    try:
        return _extract_col_ref(r_str)
    except NotColumnReferenceError:
        pass
    raise IncorrectReferenceError(
        "Argument is not a transformation or column reference."
    )



"""Tests dataloaders.py."""
from typing import cast
from unittest.mock import MagicMock

import pytest
import sqlalchemy
import torch

from bitfount.backends.pytorch.data.dataloaders import (
    DEFAULT_BUFFER_SIZE,
    PyTorchBitfountDataLoader,
    PyTorchIterableBitfountDataLoader,
)
from bitfount.backends.pytorch.data.datasets import (
    _PyTorchDataset,
    _PyTorchIterableDataset,
)
from bitfount.data.databunch import BitfountDataBunch
from bitfount.data.datasources.database_source import DatabaseSource
from bitfount.data.datastructure import DataStructure
from bitfount.data.schema import BitfountSchema
from bitfount.data.utils import DatabaseConnection
from tests.utils.helper import backend_test, integration_test, unit_test


@unit_test
@backend_test
class TestPyTorchBitfountDataLoader:
    """Tests PyTorchBitfountDataLoader."""

    def test_iterator_tab_data(self, tabular_dataset: _PyTorchDataset) -> None:
        """Tests iteration of dataloader for tabular data."""
        batch_size = 16
        dl = PyTorchBitfountDataLoader(dataset=tabular_dataset, batch_size=batch_size)
        dl_iterator = iter(dl)
        batch = next(dl_iterator)
        assert isinstance(batch, list)
        assert isinstance(batch[0], list)  # x data
        assert len(batch[0][0]) == batch_size  # tabular x data
        assert len(batch[0][1]) == batch_size  # x support columns
        assert len(batch[1]) == batch_size  # y data

    def test_iterator_image_data(self, image_dataset: _PyTorchDataset) -> None:
        """Tests iteration of dataloader for image data."""
        batch_size = 16
        dl = PyTorchBitfountDataLoader(dataset=image_dataset, batch_size=batch_size)
        dl_iterator = iter(dl)
        batch = next(dl_iterator)
        assert isinstance(batch, list)
        assert isinstance(batch[0], list)  # x data
        assert len(batch[0][0]) == batch_size  # image x data
        assert len(batch[0][1]) == batch_size  # x support columns
        assert len(batch[1]) == batch_size  # y data

    def test_iterator_image_tab_data(self, image_tab_dataset: _PyTorchDataset) -> None:
        """Tests iteration of dataloader for mixed image and tabular data."""
        batch_size = 16
        dl = PyTorchBitfountDataLoader(dataset=image_tab_dataset, batch_size=batch_size)
        dl_iterator = iter(dl)
        batch = next(dl_iterator)
        assert isinstance(batch, list)
        assert isinstance(batch[0], list)  # x data
        assert len(batch[0][0]) == batch_size  # tabular x data
        assert len(batch[0][1]) == batch_size  # image x data
        assert len(batch[0][2]) == batch_size  # x support columns
        assert len(batch[1]) == batch_size  # y data

    def test_dataloader_pytorch(self, image_tab_dataset: _PyTorchDataset) -> None:
        """Tests iteration of dataloader."""
        batch_size = 64
        dl = PyTorchBitfountDataLoader(image_tab_dataset, batch_size=batch_size)
        iterator = iter(dl)
        output = next(iterator)
        assert isinstance(output, list)
        x, y = output
        assert len(x) == 3
        assert len(x[0]) == batch_size  # tabular
        assert len(x[1]) == batch_size  # image
        assert len(x[2]) == batch_size  # support
        assert len(y) == batch_size

    def test_dataloader_non_pytorch_img_tab(
        self, image_tab_dataset: _PyTorchDataset
    ) -> None:
        """Tests x- and y-dataframe retrieval from dataloader."""
        dl = PyTorchBitfountDataLoader(image_tab_dataset)
        tab, img = dl.get_x_dataframe()
        y = dl.get_y_dataframe()
        assert len(tab) == len(y)
        assert len(img) == len(y)
        assert len(y.columns) == 1
        assert len(tab.columns) == 13
        assert len(img.columns) == 1


@backend_test
class TestPyTorchIterableBitfountDataLoader:
    """Tests PyTorchIterableBitfountDataLoader."""

    @unit_test
    @pytest.mark.parametrize(
        "dataset_length,batch_size,expected_buffer_size",
        [
            # Dataset length is the expected buffer size since it is smaller than the
            # default buffer size.
            (100, 16, 100),
            # DEFAULT_BUFFER_SIZE is the expected buffer size since it is smaller than
            # the dataset length but larger than the batch size.
            (100000, 1, DEFAULT_BUFFER_SIZE),
            # Batch size is the expected buffer size since it is larger than the default
            # buffer size.
            (100, 1024, 1024),
        ],
    )
    def test_buffer_size(
        self, batch_size: int, dataset_length: int, expected_buffer_size: int
    ) -> None:
        """Tests buffer size property calculation."""
        dataset = MagicMock(spec=_PyTorchIterableDataset)
        dataset.__len__.return_value = dataset_length
        dataloader = PyTorchIterableBitfountDataLoader(
            dataset=dataset, batch_size=batch_size
        )
        assert dataloader.buffer_size == expected_buffer_size

    @unit_test
    @pytest.mark.parametrize(
        "shuffle,secure_rng,batch_size,iterator_size",
        [
            # Tests with iterator that is smaller than the batch size
            (True, True, 8, 6),
            (False, False, 8, 7),
            # Tests with iterator that is equal to or just greater than the batch size
            (True, False, 8, 8),
            (False, False, 8, 9),
            # Tests with iterator that is significantly greater than the batch size
            # with no remainder
            (True, True, 32, 64),
            (False, False, 32, 96),
            # Tests with iterator that is significantly greater than the batch size
            # with a remainder
            (True, False, 32, 65),
            (False, False, 32, 325),
        ],
    )
    def test_iterator(
        self, batch_size: int, iterator_size: int, secure_rng: bool, shuffle: bool
    ) -> None:
        """Tests iteration of dataloader.

        The target of the batch (i.e. `batch[1]`) is used as a proxy for the rest of the
        batch for simplicity when it comes to checking shape and content.
        """
        dataset = MagicMock(spec=_PyTorchIterableDataset)
        dataset.__iter__.return_value = iter(
            [((i, i + 1), i + 2) for i in range(iterator_size)]
        )
        dataloader = PyTorchIterableBitfountDataLoader(
            dataset=dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            secure_rng=secure_rng,
        )
        for i, batch in enumerate(dataloader, start=1):
            assert isinstance(batch, list)
            if iterator_size - (i * batch_size) >= batch_size:
                assert len(batch[1]) == batch_size
            else:  # last batch may have fewer elements
                assert len(batch[1]) <= batch_size

            if shuffle:
                # There is an extremely slim chance that the order of the elements in
                # the shuffled batch will be in the same order they started in but this
                # is negligible for our batch sizes as long as it is not the final batch
                # where there may be fewer elements (e.g. just 1).
                if len(batch[1]) == batch_size:
                    assert not torch.allclose(
                        torch.sort(cast(torch.Tensor, batch[1])).values,
                        cast(torch.Tensor, batch[1]),
                    )
            else:
                assert torch.allclose(
                    torch.sort(cast(torch.Tensor, batch[1])).values,
                    cast(torch.Tensor, batch[1]),
                )

    @integration_test
    @pytest.mark.parametrize("batch_size", [4, 128])
    def test_iterator_with_multi_table_database_connection(
        self, batch_size: int, db_session: sqlalchemy.engine.base.Engine
    ) -> None:
        """Tests iteration of dataloader with underlying database connection."""
        db_conn = DatabaseConnection(db_session)
        datasource = DatabaseSource(db_conn)
        schema = BitfountSchema(
            datasource,
            force_stypes={"dummy_data": {"categorical": ["TARGET"]}},
            table_name="dummy_data",
        )
        data_structure = DataStructure(target="TARGET", table="dummy_data")

        databunch = BitfountDataBunch(
            data_structure=data_structure,
            schema=schema.get_table_schema("dummy_data"),
            datasource=datasource,
        )
        dataloader = databunch.get_train_dataloader(batch_size=batch_size)

        assert isinstance(dataloader, PyTorchIterableBitfountDataLoader)
        assert len(next(iter(dataloader))[1]) == batch_size



"""Test pod can handle incoming training requests."""
from unittest.mock import Mock, create_autospec

from grpc import RpcError
from pytest import fixture, raises

from bitfount.federated.exceptions import PodConnectFailedError
from bitfount.federated.transport.message_service import _MessageService
from bitfount.federated.transport.pod_transport import _PodMailbox
from bitfount.federated.transport.protos.messages_pb2 import SuccessResponse
from tests.utils.helper import unit_test


@unit_test
class TestPodMailbox:
    """Tests for PodMailbox."""

    @fixture
    def pod_namespace(self) -> str:
        """A pod namespace."""
        return "podOwnerUsername"

    @fixture
    def pod_name(self) -> str:
        """A pod name."""
        return "some_pod_name"

    @fixture
    def pod_identifier(self, pod_name: str, pod_namespace: str) -> str:
        """A pod identifier."""
        return f"{pod_namespace}/{pod_name}"

    @fixture
    def pod_mailbox_id(self) -> str:
        """Pod mailbox ID."""
        return "some_mailbox_id"

    @fixture
    def aes_key(self) -> bytes:
        """AES encryption key for encrypting pod message."""
        return b"aeskey_length_16"

    @fixture
    def mock_message_service(self) -> Mock:
        """Mocked message service for GRPC calls."""
        mock_message_service: Mock = create_autospec(_MessageService, instance=True)
        return mock_message_service

    @fixture
    def pod_mailbox(
        self, mock_message_service: Mock, pod_mailbox_id: str, pod_name: str
    ) -> _PodMailbox:
        """The PodMailbox under test."""
        return _PodMailbox(pod_name, pod_mailbox_id, mock_message_service)

    async def test_connect_pod_successful(
        self, mock_message_service: Mock, pod_mailbox: _PodMailbox, pod_name: str
    ) -> None:
        """connect_pod successful."""
        mock_message_service.connect_pod.return_value = SuccessResponse()
        await pod_mailbox.connect_pod(pod_name, mock_message_service)
        mock_message_service.connect_pod.assert_called_once_with(pod_name)

    async def test_connect_pod_unsuccessful(
        self,
        mock_message_service: Mock,
        pod_mailbox: _PodMailbox,
        pod_name: str,
        rpc_error: RpcError,
    ) -> None:
        """connect_pod fails."""
        mock_message_service.connect_pod.side_effect = rpc_error

        with raises(
            PodConnectFailedError,
            match=f"Failed to connect to messaging service as pod: {pod_name}",
        ):
            await pod_mailbox.connect_pod(pod_name, mock_message_service)

        mock_message_service.connect_pod.assert_called_once_with(pod_name)



"""Module containing DatabaseSource class.

DatabaseSource class handles loading data stored in a SQL database.
"""

from __future__ import annotations

import logging
from typing import Any, Dict, Iterable, List, Optional, Union, cast

import numpy as np
import pandas as pd
import sqlalchemy
from sqlalchemy import MetaData, Table, text
from sqlalchemy.exc import ArgumentError
from sqlalchemy.orm import Session

from bitfount.data.datasources.base_source import MultiTableSource
from bitfount.data.utils import (
    DatabaseConnection,
    _convert_python_dtypes_to_pandas_dtypes,
)
from bitfount.types import _Dtypes
from bitfount.utils import delegates

logger = logging.getLogger(__name__)


_DATABASE_PARTITION_SIZE: int = 100
_DATABASE_MAX_ROW_BUFFER: int = 500


@delegates()
class DatabaseSource(MultiTableSource):
    """Data source for loading data from databases."""

    def __init__(
        self,
        db_conn: Union[DatabaseConnection, str],
        database_partition_size: int = _DATABASE_PARTITION_SIZE,
        max_row_buffer: int = _DATABASE_MAX_ROW_BUFFER,
        db_conn_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)

        if isinstance(db_conn, DatabaseConnection):
            self.db_conn: DatabaseConnection = db_conn
        else:
            try:
                if not db_conn_kwargs:
                    db_conn_kwargs = {}
                self.db_conn = DatabaseConnection(db_conn, **db_conn_kwargs)
            except ArgumentError:
                url = (
                    "https://docs.sqlalchemy.org/en/14/core/engines.html#database-urls"
                )
                raise ArgumentError(
                    f"Invalid db_conn. db_conn: {db_conn} must be sqlalchemy "
                    f"compatible database url, see: {url}. "
                )
        self.database_partition_size = database_partition_size
        self.max_row_buffer = max_row_buffer
        self._con: Optional[sqlalchemy.engine.Engine] = None
        self.datastructure_query: Optional[str] = None
        self.datastructure_table_name: Optional[str] = None

    @property
    def table_names(self) -> List[str]:
        """Database table names."""
        return self.db_conn.table_names

    @property
    def query(self) -> Optional[str]:
        """A Database query as a string.

        The query is resolved in the following order:
            1. The query specified in the database connection.
            2. The table name specified in the database connection if just 1 table.
            3. The query specified by the datastructure (if multi-table).
            4. The table name specified by the datastructure (if multi-table).
            5. None.
        """
        if self.db_conn.query:
            return self.db_conn.query
        elif self.db_conn.table_names and len(self.db_conn.table_names) == 1:
            # Table name has been validated
            return f"SELECT * FROM {self.db_conn.table_names[0]}"  # nosec
        elif self.datastructure_query:
            return self.datastructure_query
        elif self.datastructure_table_name:
            # Table name has been validated
            return f"SELECT * FROM {self.datastructure_table_name}"  # nosec

        return None

    @property
    def con(self) -> sqlalchemy.engine.Engine:
        """Sqlalchemy engine.

        Connection options are set to stream results using a server side cursor where
        possible (depends on the database backend's support for this feature) with a
        maximum client side row buffer of `self.max_row_buffer` rows.
        """
        if not self._con:
            self._con = self.db_conn.con.execution_options(
                stream_results=True, max_row_buffer=self.max_row_buffer
            )
        return self._con

    def __len__(self) -> int:
        if self._data_is_loaded:
            return len(self.data)
        elif not self.db_conn.multi_table:
            data = self.get_data()
            assert data is not None  # nosec[assert_used]
            self.data = data
            return len(self.data)

        with self.con.connect() as con:
            # Ignoring the security warning because the sql query is trusted and
            # will be executed regardless.
            result = con.execute(
                text(f"SELECT COUNT(*) FROM ({self.query}) q")  # nosec
            )
            return cast(int, result.scalar_one())

    def _validate_table_name(self, table_name: str) -> None:
        """Validate the table name exists in the database.

        Args:
            table_name: The name of the table.

        Raises:
            ValueError: If the table name is not found in the data.
            ValueError: If the database connection does not have any table names.
        """
        if not self.table_names:
            raise ValueError("Database Connection is not aware of any tables.")
        elif table_name not in self.table_names:
            raise ValueError(
                f"Table name {table_name} not found in the data. "
                f"Available tables: {self.table_names}"
            )

    def get_values(
        self, col_names: List[str], table_name: Optional[str] = None, **kwargs: Any
    ) -> Dict[str, Iterable[Any]]:
        """Get distinct values from columns in Database dataset.

        Args:
            col_names: The list of the columns whose distinct values should be
                returned.
            table_name: The name of the table to which the column exists. Required
                for multi-table databases.

        Returns:
            The distinct values of the requested column as a mapping from col name to
            a series of distinct values.
        """
        metadata = MetaData(self.con)
        output: Dict[str, Iterable[Any]] = {}
        if self.query is not None:
            # TODO: [BIT-1595] change to load memory using sqlalchemy FrozenResult
            self.data = cast(pd.DataFrame, pd.read_sql(self.query, con=self.con))
            for col_name in col_names:
                output[col_name] = self.data[col_name].unique()
        elif table_name is not None:
            self._validate_table_name(table_name)

            table = Table(
                table_name,
                metadata,
                schema=self.db_conn.db_schema,
                autoload=True,
                autoload_with=self.con,
            )
            with Session(self.con) as session:
                for col_name in col_names:
                    values = np.array(
                        [v for v, in session.query(table.columns[col_name]).distinct()]
                    )
                    output[col_name] = values
        else:
            raise ValueError("No table name provided for multi-table datasource.")
        return output

    def get_column(
        self, col_name: str, table_name: Optional[str] = None, **kwargs: Any
    ) -> pd.Series:
        """Loads and returns single column from Database dataset.

        Args:
            col_name: The name of the column which should be loaded.
            table_name: The name of the table to which the column exists. Required
                for multi-table databases.

        Returns:
            The column requested as a series.

        Raises:
            ValueError: If the data is multi-table but no table name provided.
            ValueError: If the table name is not found in the data.
        """
        results: Iterable[Any]
        metadata = MetaData(self.con)
        if self.query is not None:
            with Session(self.con) as session:
                results = session.execute(text(self.query)).columns(col_name)
        elif table_name is not None:
            self._validate_table_name(table_name)

            table = Table(
                table_name,
                metadata,
                schema=self.db_conn.db_schema,
                autoload=True,
                autoload_with=self.con,
            )
            with Session(self.con) as session:
                results = session.query(table.columns[col_name])
        else:
            raise ValueError("No table name provided for multi-table datasource.")
        series = pd.Series([v for v, in results])
        return series

    def get_data(
        self,
        sql_query: Optional[str] = None,
        table_name: Optional[str] = None,
        **kwargs: Any,
    ) -> Optional[pd.DataFrame]:
        """Loads and returns data from Database dataset.

        Args:
            sql_query: A SQL query string required for multi table data sources. This
                comes from the DataStructure and takes precedence over the table_name.
            table_name: Table name for multi table data sources. This
                comes from the DataStructure and is ignored if sql_query has been
                provided.

        Returns:
            A DataFrame-type object which contains the data.
        """
        data: Optional[pd.DataFrame] = None
        if self.db_conn.multi_table:
            # If multi-table, we don't actually load the data, we just set the query and
            # table name for the iterator.
            if sql_query is not None:
                self.datastructure_query = sql_query
            if table_name is not None:
                self.datastructure_table_name = table_name
        elif self.db_conn.query:
            data = pd.read_sql_query(sql=self.db_conn.query, con=self.con)
        else:
            # If the data is not multi-table and there is no query, there must
            # necessarily be one table name. Reassuring mypy of this.
            assert (  # nosec[assert_used]
                self.table_names is not None and len(self.table_names) == 1
            )
            query = f"SELECT * FROM {self.table_names[0]}"  # nosec
            data = pd.read_sql_query(
                sql=query,
                con=self.con,
            )
        return data

    def get_dtypes(self, table_name: Optional[str] = None, **kwargs: Any) -> _Dtypes:
        """Loads and returns the columns and column types from the Database dataset.

        Args:
            table_name: The name of the table_name which should be loaded. Only
                required for multitable database.

        Returns:
            A mapping from column names to column types.

        Raises:
            ValueError: If the data is multi-table but no table name provided.

        """
        metadata = MetaData(self.con)
        dtypes: _Dtypes
        if self.query is not None:
            with Session(self.con) as session:
                result = session.execute(text(self.query)).first()
            data = pd.DataFrame([result])
            dtypes = self._get_data_dtypes(data)

        elif table_name is not None:
            self._validate_table_name(table_name)

            table = Table(
                table_name,
                metadata,
                schema=self.db_conn.db_schema,
                autoload=True,
                autoload_with=self.con,
            )
            dtypes = {
                col.name: _convert_python_dtypes_to_pandas_dtypes(
                    col.type.python_type, col.name
                )
                for col in table.columns
            }
        else:
            raise ValueError("No table name provided for multi-table datasource.")
        return dtypes

    @property
    def multi_table(self) -> bool:
        """Attribute to specify whether the datasource is multi table."""
        return self.db_conn.multi_table



"""Unit tests for _PodVitals webserver."""
import time
from unittest.mock import AsyncMock, Mock

from aiohttp.test_utils import TestClient, TestServer
import pytest
from pytest import MonkeyPatch
from pytest_mock import MockerFixture

from bitfount.federated.pod_vitals import _PodVitals, _PodVitalsHandler
from tests.utils.helper import unit_test
from tests.utils.mocks import create_dataclass_mock


@unit_test
class TestPodVitals:
    """Test TestPodVitals."""

    async def test__last_task_execution_time_update(self) -> None:
        """Test last_task_execution_time is updated as expected."""
        pod_vitals = _PodVitals()
        new_time = time.time()
        lock_acquired_mock = Mock()
        lock_released_mock = Mock()
        pod_vitals._last_task_execution_lock = Mock(
            __enter__=lock_acquired_mock, __exit__=lock_released_mock
        )

        pod_vitals.last_task_execution_time = new_time

        # check lock was acquired in updating last_task_execution_time
        lock_acquired_mock.assert_called_once()
        lock_released_mock.assert_called_once()
        assert pod_vitals.last_task_execution_time == new_time
        # check lock was acquired again in above assertion when
        # retrieving last_task_execution_time
        assert lock_acquired_mock.call_count == 2
        assert lock_acquired_mock.call_count == 2


@unit_test
class Test_PodVitalsHandler:
    """Test Test_PodVitalsHandler."""

    @pytest.fixture
    def pod_vitals(self) -> Mock:
        """Pod vitals mock fixture."""
        mock_pod_vitals = create_dataclass_mock(_PodVitals)
        mock_pod_vitals.last_task_execution_time = time.time()
        return mock_pod_vitals

    @pytest.fixture
    def handler(self, pod_vitals: _PodVitals) -> _PodVitalsHandler:
        """Pod vitals handler fixture."""
        handler = _PodVitalsHandler(pod_vitals)
        handler.runner = Mock()
        return handler

    @pytest.fixture
    def server(self, handler: _PodVitalsHandler) -> TestServer:
        """Test server fixture."""
        return TestServer(handler.app)

    @pytest.fixture
    def mock_open_socket(self, mocker: MockerFixture) -> AsyncMock:
        """Mocks out the Modeller.run() method in protocol.py."""
        mock_open_socket_method: AsyncMock = mocker.patch(
            "bitfount.federated.pod_vitals._PodVitalsHandler._open_socket"
        )
        mock_open_socket_method.return_value = 8080
        return mock_open_socket_method

    async def test__status_request(self, server: TestServer) -> None:
        """Test /status endpoint to determines if the webserver is responding."""
        async with TestClient(server) as client:
            resp = await client.request("GET", "/status")
            json_resp = await resp.json()
            assert resp.status == 200
            assert json_resp == {"status": "OK"}

    async def test__health_request(self, server: TestServer) -> None:
        """Test /health endpoint determines if a pod is healthy."""
        async with TestClient(server) as client:
            resp = await client.request("GET", "/health")
            json_resp = await resp.json()
            assert resp.status == 200
            assert json_resp == {"healthy": True}

    async def test__unhealth_request(
        self, mocker: MockerFixture, server: TestServer
    ) -> None:
        """Test /health endpoint determines if a pod is unhealthy.

        If the last time a task was executed is greate than
        MAX_TASK_EXECUTION_TIME, the endpoint should return that
        the pod is unhealthly.
        """
        mocker.patch("bitfount.federated.pod_vitals.MAX_TASK_EXECUTION_TIME", 0)
        async with TestClient(server) as client:
            resp = await client.request("GET", "/health")
            json_resp = await resp.json()
            assert resp.status == 200
            assert json_resp == {"healthy": False}

    async def test__open_socket(
        self, handler: _PodVitalsHandler, mocker: MockerFixture
    ) -> None:
        """Test the socket library is used to get an open port."""
        expected_port = 8080
        mock_socket = mocker.patch("socket.socket")
        mock_socket.return_value.recv.return_value = 1
        mock_socket.return_value.getsockname.return_value = ("host", expected_port)
        port = handler._open_socket()
        assert port == expected_port
        mock_socket.return_value.getsockname.assert_called_once()

    async def test__pod_vitals_port_env_var_set(
        self,
        handler: _PodVitalsHandler,
        mock_open_socket: Mock,
        monkeypatch: MonkeyPatch,
    ) -> None:
        """Test pod health port is set to BITFOUNT_POD_VITALS_PORT env var."""
        pod_vitals_port = 8080
        monkeypatch.setenv("BITFOUNT_POD_VITALS_PORT", str(pod_vitals_port))
        result = handler._get_pod_vitals_port()
        assert result == pod_vitals_port
        assert not mock_open_socket.called

    async def test__pod_vitals_port_no_env_var(
        self, handler: _PodVitalsHandler, mock_open_socket: Mock
    ) -> None:
        """Test pod health port when BITFOUNT_POD_VITALS_PORT not set.

        If BITFOUNT_POD_VITALS_PORT is not set as an environment variable
        the port number should be determined by _open_socket method.
        """
        port = 8080
        result = handler._get_pod_vitals_port()
        assert result == port
        mock_open_socket.assert_called_once()

    async def test__pod_vitals_error_invalid_port(
        self,
        handler: _PodVitalsHandler,
        mock_open_socket: Mock,
        monkeypatch: MonkeyPatch,
    ) -> None:
        """Test exception raised when BITFOUNT_POD_VITALS_PORT is invalid."""
        pod_vitals_port = "abc123"
        expected_error_msg = (
            "BITFOUNT_POD_VITALS_PORT must be an integer. "
            f"BITFOUNT_POD_VITALS_PORT set to '{pod_vitals_port}'"
        )
        monkeypatch.setenv("BITFOUNT_POD_VITALS_PORT", str(pod_vitals_port))
        with pytest.raises(ValueError) as error:
            handler._get_pod_vitals_port()
            assert str(error.value) == expected_error_msg

        assert not mock_open_socket.called

    async def test__pod_vitals_error_invalid_port_range(
        self,
        handler: _PodVitalsHandler,
        mock_open_socket: Mock,
        monkeypatch: MonkeyPatch,
    ) -> None:
        """Test exception raised when BITFOUNT_POD_VITALS_PORT is invalid."""
        pod_vitals_port = "99999"
        monkeypatch.setenv("BITFOUNT_POD_VITALS_PORT", str(pod_vitals_port))
        with pytest.raises(ValueError) as error:
            handler._get_pod_vitals_port()
            assert (
                str(error.value)
                == "Invalid BITFOUNT_POD_VITALS_PORT given. Must be in range [1-65535]"
            )
        assert not mock_open_socket.called

    async def test__start_webserver(
        self,
        handler: _PodVitalsHandler,
        mocker: MockerFixture,
        monkeypatch: MonkeyPatch,
    ) -> None:
        """Ensure server is started with expected settings."""
        port = 8080
        monkeypatch.setenv("BITFOUNT_POD_VITALS_PORT", str(port))
        mock_tcp_site = mocker.patch("bitfount.federated.pod_vitals.TCPSite")
        mock_loop = AsyncMock()

        handler.start(mock_loop)

        mock_tcp_site.assert_called_once_with(handler.runner, "0.0.0.0", port)
        mock_tcp_site.return_value.start.assert_called_once()



"""Fixtures related to environment setup, envvars, etc."""
from typing import Iterator

from _pytest.monkeypatch import MonkeyPatch
from pytest import fixture

import bitfount
from tests.utils import PytestRequest


@fixture(autouse=True)
def env_fix(monkeypatch: MonkeyPatch) -> None:
    """Fix the environment into a known state for tests."""
    # Sets the environment and engine to use the BasicDataFactory. If a specific
    # engine is needed (aka PyTorch), this must be overridden in a fixture of the
    # same name in a conftest.py file closer to the test.
    monkeypatch.setenv("BITFOUNT_ENGINE", bitfount.config._BASIC_ENGINE)
    monkeypatch.setattr(
        "bitfount.config.BITFOUNT_ENGINE", bitfount.config._BASIC_ENGINE
    )


@fixture
def environment(monkeypatch: MonkeyPatch, request: PytestRequest) -> None:
    """Sets up the BITFOUNT_ENVIRONMENT environment variable."""
    environment = request.param
    if environment:
        monkeypatch.setenv("BITFOUNT_ENVIRONMENT", environment)


@fixture(autouse=True)
def cache_clear() -> Iterator[None]:
    """Clears the cache of get_environment before and after each test."""
    bitfount.config._get_environment.cache_clear()
    yield
    bitfount.config._get_environment.cache_clear()



"""MixIn classes for compatible models with the federated algorithms."""
from __future__ import annotations

from abc import ABC, abstractmethod
import os
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
import numpy as np
import pandas as pd

from bitfount.data.dataloaders import BitfountDataLoader
from bitfount.data.datasources.base_source import BaseSource
from bitfount.federated.aggregators.base import _BaseAggregatorFactory
from bitfount.federated.algorithms.model_algorithms.evaluate import ModelEvaluation
from bitfount.federated.algorithms.model_algorithms.federated_training import (
    FederatedModelTraining,
)
from bitfount.federated.authorisation_checkers import IdentityVerificationMethod
from bitfount.federated.exceptions import PSIMultiplePodsError
from bitfount.federated.helper import (
    _check_and_update_pod_ids,
    _create_aggregator,
    _create_message_service,
    _get_idp_url,
)
from bitfount.federated.modeller import _Modeller
from bitfount.federated.protocols.model_protocols.federated_averaging import (
    FederatedAveraging,
)
from bitfount.federated.protocols.psi import (
    PrivateSetIntersection,
    _PSICompatibleAlgoFactory_,
)
from bitfount.federated.protocols.results_only import (
    ResultsOnly,
    _ResultsOnlyCompatibleAlgoFactory_,
)
from bitfount.federated.transport.message_service import _MessageService
from bitfount.hub.helper import _create_bitfounthub

if TYPE_CHECKING:
    from bitfount.data.datastructure import DataStructure
    from bitfount.federated.transport.config import MessageServiceConfig
    from bitfount.hub.api import BitfountHub
    from bitfount.metrics import Metric
    from bitfount.types import (
        T_DTYPE,
        _SerializedWeights,
        _Weights,
        _Residuals,
        _DistributedModelTypeOrReference,
    )

from bitfount.federated.logging import _get_federated_logger

logger = _get_federated_logger(__name__)


class _DistributedModelMixIn(ABC):
    """A mixin for models used in federated mechanisms.

    An abstract base mixin for models that are compatible with the following
    distributed learning protocols:
        - FederatedAveraging
    """

    datastructure: DataStructure
    # Denotes the Pod the model is running in (if any)
    pod_identifier: Optional[str] = None

    @abstractmethod
    def get_param_states(self) -> _Weights:
        """Gets the current states of the trainable parameters of the model.

        Returns:
            A dict of param names to tensors
        """
        raise NotImplementedError

    @abstractmethod
    def apply_weight_updates(self, weight_updates: Sequence[_Residuals]) -> _Weights:
        """Applies weight updates to the weights of this model.

        Apply a sequence of parameter weight updates (mappings of parameter name
        to a tensor describing the weight update) to the parameters of this model.
        Used by Modeller to apply updates received from Workers.

        Args:
            weight_updates (Sequence[_Residuals]): The sequence
                of weight updates

        Returns:
            The updated parameters as a dict of name to tensor.
        """
        raise NotImplementedError

    @abstractmethod
    def update_params(self, new_model_params: _Weights) -> None:
        """Updates the current model parameters to the ones provided.

        Used by Worker to update new parameters received from Modeller.

        Args:
            new_model_params (WeightMapping): The new model parameters
                to update to as a mapping of parameter names to tensors.
        """
        raise NotImplementedError

    @abstractmethod
    def deserialize_params(self, serialized_weights: _SerializedWeights) -> _Weights:
        """Converts serialized model parameters to tensors.

        Used by Worker to convert serialized parameters received from Modeller.

        Args:
            serialized_weights (_SerializedWeights): The model parameters
                to deserialize.
        """
        raise NotImplementedError

    @abstractmethod
    def serialize_params(self, weights: _Weights) -> _SerializedWeights:
        """Serializes model parameters.

        Used by Modeller to serialize model parameters to send to the Worker.

        Args:
            weights (_Weights): The model parameters
                to serialize.
        """
        raise NotImplementedError

    @abstractmethod
    def diff_params(self, old_params: _Weights, new_params: _Weights) -> _Residuals:
        """Calculates the difference between two sets of model parameters."""
        raise NotImplementedError

    @abstractmethod
    def set_model_training_iterations(self, iterations: int) -> None:
        """Sets model steps or epochs to the appropriate number between updates."""
        raise NotImplementedError

    @abstractmethod
    def reset_trainer(self) -> None:
        """Resets the trainer to its initial state.

        :::note

        Importantly, calling this method in between `fit` calls allows the caller to
        repeatedly refit the model continuing from the batch after the one that was last
        fit. This only applies to step-wise training.

        :::
        """
        raise NotImplementedError

    @abstractmethod
    def log_(self, name: str, value: Any, **kwargs: Any) -> Any:
        """Logs a metric with a particular value to the user's configured model loggers.

        Args:
            name: The name of the metric to log
            value: The value of the metric to log
            **kwargs: Additional keyword arguments to pass to the logger
        """
        raise NotImplementedError

    @abstractmethod
    def tensor_precision(self) -> T_DTYPE:
        """Gets the floating point precision used by model tensors.

        Typically this will be 32 bits.
        """
        raise NotImplementedError

    @abstractmethod
    def _fit_local(
        self,
        data: BaseSource,
        metrics: Optional[MutableMapping[str, Metric]] = None,
        **kwargs: Any,
    ) -> Dict[str, str]:
        """Fits model locally using data from `datasource`.

        Should be implemented in the final model class that subclasses
        DistributedModelMixIn.
        """
        raise NotImplementedError

    @abstractmethod
    def _get_model(
        self,
        extra_imports: Optional[List[str]] = None,
        hub: Optional[BitfountHub] = None,
        **kwargs: Any,
    ) -> _DistributedModelTypeOrReference:
        """Gets the model to be used in distributed training.

        This method is used by the Modeller to get the model to be used in
        distributed training.

        Args:
            extra_imports: A list of extra imports to add to the model file if the model
                is a custom model.
            hub: A BitfountHub instance to be passed to `BitfountModelReference`.
                Optional.
            **kwargs: Additional keyword arguments to pass to the
                `BitfountModelReference` constructor as hyperparameters.

        Returns:
            Either a `BitfountModelReference` object if the model is a custom model, or
            the model itself if it is a built-in model.
        """
        raise NotImplementedError

    def set_pod_identifier(self, pod_identifier: str) -> None:
        """Sets the pod identifier for the model.

        This must be called on the Pod/Worker side in Distributed training because it
        is needed for the model to be able to extract the relevant information from the
        datastructure sent by the Modeller.

        Args:
            pod_identifier: The pod identifier for the model.
        """
        self.pod_identifier = pod_identifier

    def _fit_federated(
        self,
        pod_identifiers: List[str],
        username: Optional[str] = None,
        aggregator: Optional[_BaseAggregatorFactory] = None,
        steps_between_parameter_updates: Optional[int] = None,
        epochs_between_parameter_updates: Optional[int] = None,
        bitfounthub: Optional[BitfountHub] = None,
        ms_config: Optional[MessageServiceConfig] = None,
        message_service: Optional[_MessageService] = None,
        pod_public_key_paths: Optional[Mapping[str, Path]] = None,
        pretrained_file: Optional[Union[str, os.PathLike]] = None,
        secure_aggregation: bool = False,
        auto_eval: bool = True,
        identity_verification_method: IdentityVerificationMethod = IdentityVerificationMethod.DEFAULT,  # noqa: B950
        private_key_or_file: Optional[Union[RSAPrivateKey, Path]] = None,
        idp_url: Optional[str] = None,
        model_out: Optional[Union[Path, str]] = None,
        require_all_pods: bool = False,
        extra_imports: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Optional[Dict[str, str]]:
        """Fits model federated-ly."""
        if not bitfounthub:
            bitfounthub = _create_bitfounthub(username=username)

        model = self._get_model(extra_imports=extra_imports, hub=bitfounthub, **kwargs)
        algorithm = FederatedModelTraining(model=model, pretrained_file=pretrained_file)

        pod_identifiers = _check_and_update_pod_ids(pod_identifiers, bitfounthub)
        if (
            len(pod_identifiers) > 1
            and hasattr(self.datastructure, "table")
            and isinstance(self.datastructure.table, str)
        ):
            raise ValueError(
                "You are attempting to train on multiple pods, and the"
                "provided the DataStructure table name is a string. "
                "Please make sure that the `table` argument to the "
                "`DataStructure` is a mapping of Pod names to table names. "
            )
        datastructure_pod_identifiers = self.datastructure.get_pod_identifiers()
        if datastructure_pod_identifiers:
            datastructure_pod_identifiers = _check_and_update_pod_ids(
                datastructure_pod_identifiers, bitfounthub
            )
            self.datastructure._update_datastructure_with_hub_identifiers(
                datastructure_pod_identifiers
            )
        if not message_service:
            message_service = _create_message_service(bitfounthub.session, ms_config)

        if not aggregator:
            aggregator = _create_aggregator(secure_aggregation=secure_aggregation)

        if not idp_url:
            idp_url = _get_idp_url()

        # TODO: [BIT-1098] Manage pods with different schemas
        protocol = FederatedAveraging(
            algorithm=algorithm,
            aggregator=aggregator,
            steps_between_parameter_updates=steps_between_parameter_updates,
            epochs_between_parameter_updates=epochs_between_parameter_updates,
            auto_eval=auto_eval,
        )

        modeller = _Modeller(
            protocol=protocol,
            message_service=message_service,
            bitfounthub=bitfounthub,
            pod_public_key_paths=pod_public_key_paths,
            identity_verification_method=identity_verification_method,
            private_key=private_key_or_file,
            idp_url=idp_url,
        )

        # Run task
        result = modeller.run(
            pod_identifiers, require_all_pods=require_all_pods, model_out=model_out
        )

        if result is False:
            return None
        else:
            return result

    def fit(
        self,
        data: Optional[BaseSource] = None,
        metrics: Optional[Dict[str, Metric]] = None,
        pod_identifiers: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Optional[Dict[str, str]]:
        """Fits model either locally or federated-ly.

        `pod_identifiers` must be provided for federated training.

        :::caution

        For remote training, this method does not use the existing parameters of the
        model unless they are serialized and provided using the `pretrained_file`
        keyword argument.

        :::

        Args:
            data: Datasource for training. Defaults to None.
            metrics: Metrics to calculate for validation. Defaults to None.
            pod_identifiers: List of pod identifiers. If this is provided, the model
                will be trained in a federated manner. Defaults to None.
            **kwargs: Optional keyword arguments passed to the federated fit method.
                Any unrecognized keyword arguments will be interpreted as
                custom model hyperparameters.

        Returns:
            A dictionary of metrics and their values. Optional.

        Raises:
            ValueError: If neither `pod_identifiers` are provided for federated training
                nor `data` is provided for local training.
        """
        if pod_identifiers is not None:
            logger.info(f"Training federated with pods: {', '.join(pod_identifiers)}.")
            return self._fit_federated(pod_identifiers=pod_identifiers, **kwargs)
        elif data is not None:
            logger.info("Training locally using provided datasource.")
            return self._fit_local(data=data, metrics=metrics)

        raise ValueError(
            "Please provide either pod identifiers "
            "or a datasource for model training."
        )

    def evaluate(
        self,
        test_dl: Optional[BitfountDataLoader] = None,
        pod_identifiers: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Union[Tuple[np.ndarray, np.ndarray], Dict[str, float]]:
        """Evaluates model either locally or federated-ly.

        `pod_identifiers` must be provided for federated evaluation.

        :::caution

        For remote evaluation, this method does not use the existing parameters of the
        model unless they are serialized and provided using the `pretrained_file`
        keyword argument.

        :::

        Args:
            test_dl: Optional dataloader to run inference on which takes precedence over
                the dataloader returned by `self.test_dataloader`.
            pod_identifiers: List of pod identifiers. If this is provided, the model
                will be trained in a federated manner. Defaults to None.
            **kwargs: Optional keyword arguments passed to the federated evaluate
                method.

        Returns:
            Either:
                - A tuple of numpy arrays containing the predicted and actual values if
                evaluated locally.
                - A dictionary of metrics if evaluated remotely.
        """
        if pod_identifiers is not None:
            logger.info(f"Evaluating model on pods: {', '.join(pod_identifiers)}.")
            return self._evaluate_federated(pod_identifiers=pod_identifiers, **kwargs)

        logger.info("Evaluating model locally")
        return self._evaluate_local(test_dl=test_dl)

    def _evaluate_local(
        self, test_dl: Optional[BitfountDataLoader] = None, **kwargs: Any
    ) -> Tuple[np.ndarray, np.ndarray]:
        """This method runs inference on the test dataloader.

        Args:
            test_dl: Optional dataloader to run inference on which takes precedence over
                the dataloader returned by `self.test_dataloader`.

        Returns:
            A tuple of predictions and targets as numpy arrays.
        """
        raise NotImplementedError

    def _evaluate_federated(
        self,
        pod_identifiers: List[str],
        username: Optional[str] = None,
        aggregator: Optional[_BaseAggregatorFactory] = None,
        bitfounthub: Optional[BitfountHub] = None,
        ms_config: Optional[MessageServiceConfig] = None,
        message_service: Optional[_MessageService] = None,
        pod_public_key_paths: Optional[Mapping[str, Path]] = None,
        pretrained_file: Optional[Union[str, os.PathLike]] = None,
        secure_aggregation: bool = False,
        identity_verification_method: IdentityVerificationMethod = IdentityVerificationMethod.DEFAULT,  # noqa: B950
        private_key_or_file: Optional[Union[RSAPrivateKey, Path]] = None,
        idp_url: Optional[str] = None,
        model_out: Optional[Union[Path, str]] = None,
        require_all_pods: bool = False,
        extra_imports: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Dict[str, float]:
        """Evaluates model remotely on all the specified pods separately.

        Returns:
            A dictionary of metrics.
        """
        if not bitfounthub:
            bitfounthub = _create_bitfounthub(username=username)

        model = self._get_model(extra_imports=extra_imports, hub=bitfounthub, **kwargs)
        algorithm = ModelEvaluation(model=model, pretrained_file=pretrained_file)

        pod_identifiers = _check_and_update_pod_ids(pod_identifiers, bitfounthub)
        if (
            len(pod_identifiers) > 1
            and hasattr(self.datastructure, "table")
            and isinstance(self.datastructure.table, str)
        ):
            raise ValueError(
                "You are attempting to train on multiple pods, and the"
                "provided the DataStructure table name is a string. "
                "Please make sure that the `table` argument to the "
                "`DataStructure` is a mapping of Pod names to table names. "
            )
        datastructure_pod_identifiers = self.datastructure.get_pod_identifiers()
        if datastructure_pod_identifiers:
            datastructure_pod_identifiers = _check_and_update_pod_ids(
                datastructure_pod_identifiers, bitfounthub
            )
            self.datastructure._update_datastructure_with_hub_identifiers(
                datastructure_pod_identifiers
            )
        if not message_service:
            message_service = _create_message_service(bitfounthub.session, ms_config)

        if not idp_url:
            idp_url = _get_idp_url()

        protocol = ResultsOnly(
            algorithm=algorithm,
            aggregator=aggregator,
            secure_aggregation=secure_aggregation,
        )

        modeller = _Modeller(
            protocol=protocol,
            message_service=message_service,
            bitfounthub=bitfounthub,
            pod_public_key_paths=pod_public_key_paths,
            identity_verification_method=identity_verification_method,
            private_key=private_key_or_file,
            idp_url=idp_url,
        )

        metrics = modeller.run(
            pod_identifiers, require_all_pods=require_all_pods, model_out=model_out
        )
        return cast(Dict[str, float], metrics)


class _ModellessAlgorithmMixIn(ABC):
    """A mixin for SQL Algorithms.

    An abstract base mixin for SQL algorithms that are compatible with the
    ResultOnly protocol.
    """

    def execute(
        self,
        pod_identifiers: List[str],
        username: Optional[str] = None,
        bitfounthub: Optional[BitfountHub] = None,
        ms_config: Optional[MessageServiceConfig] = None,
        message_service: Optional[_MessageService] = None,
        pod_public_key_paths: Optional[Mapping[str, Path]] = None,
        identity_verification_method: IdentityVerificationMethod = IdentityVerificationMethod.DEFAULT,  # noqa: B950
        private_key_or_file: Optional[Union[RSAPrivateKey, Path]] = None,
        idp_url: Optional[str] = None,
        require_all_pods: bool = False,
        aggregator: Optional[_BaseAggregatorFactory] = None,
    ) -> List[pd.DataFrame]:
        """Execute ResultsOnly compatible algorithm.

        Syntactic sugar to allow the modeller to call `.execute(...)` on
        ResultsOnly compatible algorithms.
        """
        if not bitfounthub:
            bitfounthub = _create_bitfounthub(username=username)

        pod_identifiers = _check_and_update_pod_ids(pod_identifiers, bitfounthub)

        if not message_service:
            message_service = _create_message_service(bitfounthub.session, ms_config)

        if not idp_url:
            idp_url = _get_idp_url()

        protocol = ResultsOnly(
            algorithm=cast(_ResultsOnlyCompatibleAlgoFactory_, self),
            aggregator=aggregator,
        )
        modeller = _Modeller(
            protocol=protocol,
            message_service=message_service,
            bitfounthub=bitfounthub,
            pod_public_key_paths=pod_public_key_paths,
            identity_verification_method=identity_verification_method,
            private_key=private_key_or_file,
            idp_url=idp_url,
        )

        result = cast(
            List[pd.DataFrame],
            modeller.run(
                pod_identifiers=pod_identifiers, require_all_pods=require_all_pods
            ),
        )
        return result


class _PSIAlgorithmsMixIn(ABC):
    """A mixin for PSI Algorithms.

    An abstract base mixin for algorithms that are compatible with the
    PSI protocol.
    """

    def execute(
        self: _PSICompatibleAlgoFactory_,
        pod_identifiers: List[str],
        datasource: BaseSource,
        datasource_columns: Optional[List[str]] = None,
        datasource_table: Optional[str] = None,
        pod_columns: Optional[List[str]] = None,
        pod_table: Optional[str] = None,
        username: Optional[str] = None,
        bitfounthub: Optional[BitfountHub] = None,
        ms_config: Optional[MessageServiceConfig] = None,
        message_service: Optional[_MessageService] = None,
        pod_public_key_paths: Optional[Mapping[str, Path]] = None,
        identity_verification_method: IdentityVerificationMethod = IdentityVerificationMethod.DEFAULT,  # noqa: B950
        private_key_or_file: Optional[Union[RSAPrivateKey, Path]] = None,
        idp_url: Optional[str] = None,
        require_all_pods: bool = False,
    ) -> List[pd.DataFrame]:
        """Execute PSI compatible algorithm.

        Syntactic sugar to allow the modeller to call `.intersect(...)` on
        PrivateSetIntersection compatible algorithms.

        Args:
            pod_identifiers: The pod identifier for the Private Set
                Intersection as a list.
            datasource: The modeller datasource on which the Private
                Set Intersection will be performed on.
            datasource_columns: The modeller's columns from their datasource
                on which the private set intersection will be computed as a
                list of strings. Defaults to None.
            datasource_table: The modeller's table from their datasource,
                if the datasource is multitable, on which the private set
                intersection will be computed as a string. Defaults to None.
            pod_columns: The pod's columns from their datasource
                on which the private set intersection will be computed as a
                list of strings. Defaults to None.
            pod_table: The pod's table from their datasource,
                if the datasource is multitable, on which the private set
                intersection will be computed as a string. Defaults to None.
            username: The modeller's username. Defaults to None.
            bitfounthub: The bifount hub instance. Defaults to None.
            ms_config: Configuration for the message service. Defaults to None.
            message_service: The message service to use. Defaults to None.
            pod_public_key_paths: The path for the pod public key. Used
                when authentication with the pod is done with the public key.
                Defaults to None.
            identity_verification_method: The identity verification method
                to use. Defaults to None.
            private_key_or_file: The private key or path to the private
                key. Defaults to None.
            idp_url: The identity provider url. Defaults to None.
            require_all_pods: Whether all pods are needed for the algorithm.
                Defaults to False.

        Returns:
            The records from the modeller dataset which were found in the intersection.

        Raises:
            PSIMultiplePodsError: If `execute` is called on multiple pods.
        """
        if len(pod_identifiers) > 1:
            raise PSIMultiplePodsError(
                "You are trying to run a PrivateSetIntersection "
                "task on multiple pods. We support "
                "PrivateSetIntersection on a single pod."
            )
        if not bitfounthub:
            bitfounthub = _create_bitfounthub(username=username)

        pod_identifiers = _check_and_update_pod_ids(pod_identifiers, bitfounthub)

        if not message_service:
            message_service = _create_message_service(bitfounthub.session, ms_config)

        if not idp_url:
            idp_url = _get_idp_url()

        protocol = PrivateSetIntersection(
            algorithm=self,
            datasource=datasource,
            datasource_columns=datasource_columns,
            datasource_table=datasource_table,
            pod_columns=pod_columns,
            pod_table=pod_table,
        )
        modeller = _Modeller(
            protocol=protocol,
            message_service=message_service,
            bitfounthub=bitfounthub,
            pod_public_key_paths=pod_public_key_paths,
            identity_verification_method=identity_verification_method,
            private_key=private_key_or_file,
            idp_url=idp_url,
        )
        result = cast(
            List[pd.DataFrame],
            modeller.run(
                pod_identifiers=pod_identifiers, require_all_pods=require_all_pods
            ),
        )
        return result
